/*
 * CoordinateSystem.cpp
 *
 *  Created on: Oct 2, 2017
 *      Author: julien
 */

#include "coordinatesystem.hpp"

Coordinate_System::Coordinate_System(int system_length,
									 double length_ratio,
									 int marking_rate)
								   : coordinate_length(system_length),
									 range(system_length/2),
									 length_ratio(length_ratio),
									 marking_rate(marking_rate)
{
	// create mat coordinate system
	system = Mat(coordinate_length, coordinate_length, CV_8UC3, Scalar(0));
	// define the color
	Scalar axis_color(180, 180, 180);
	Scalar line_color(80, 80, 80);


	// draw the axes
	line(system, Point(0, range), Point(system_length, range), axis_color, 1, 8, 0);
	line(system, Point(range, 0), Point(range, system_length), axis_color, 1, 8, 0);

	// draw the markings
	int i = marking_rate;
	while(i <= range/length_ratio){

		line(system, Point(range - i * length_ratio, 0),
					 Point(range - i * length_ratio, system_length),
					 line_color, 1, 8, 0);

		line(system, Point(0, range - i * length_ratio),
					 Point(system_length, range - i * length_ratio),
					 line_color, 1, 8, 0);

		line(system, Point(range + i * length_ratio, 0),
					 Point(range + i * length_ratio, system_length),
					 line_color, 1, 8, 0);

		line(system, Point(0, range + i * length_ratio),
					 Point(system_length, range + i * length_ratio),
					 line_color, 1, 8, 0);

		line(system, Point(range - i * length_ratio, range - 4),
					 Point(range - i * length_ratio, range + 4),
					 axis_color, 1, 8, 0);

		line(system, Point(range - 4, range - i * length_ratio),
					 Point(range + 4, range - i * length_ratio),
					 axis_color, 1, 8, 0);

		line(system, Point(range + i * length_ratio, range - 4),
					 Point(range + i * length_ratio, range + 4),
					 axis_color, 1, 8, 0);

		line(system, Point(range - 4, range + i * length_ratio),
					 Point(range + 4, range + i * length_ratio),
					 axis_color, 1, 8, 0);

		i += marking_rate;
	}

	// label axes
	#ifdef SWITCHED_AXIS
		putText(system, "x", Point(range + 10, 12),
				FONT_HERSHEY_SIMPLEX, 0.5, axis_color, 1, 8);
		putText(system, "y", Point(4, range - 10),
				FONT_HERSHEY_SIMPLEX, 0.5, axis_color, 1, 8);

		putText(system, to_string(marking_rate),
				Point(range + 10, range - length_ratio * marking_rate + 4),
				FONT_HERSHEY_SIMPLEX, 0.5, axis_color, 1, 8);

	#else
		putText(system, "y", Point(range + 18, 18),
				FONT_HERSHEY_SIMPLEX, 0.8, axis_color, 1, 8);
		putText(system, "x", Point(range - 18, system_length - 15),
				FONT_HERSHEY_SIMPLEX, 0.8, axis_color, 1, 8);
	#endif

}

Function::pPoint Coordinate_System::pixel2central_coord(Point pixel_point){

	// look at the documentation to see the reasoning behind this calculation
	// TODO docu

	#ifdef SWITCHED_AXIS
		return Function::pPoint((range - pixel_point.y)/length_ratio,
					 (range - pixel_point.x)/length_ratio);
	#else
		return Point((range - pixel_point.x)/length_ratio,
				 	 (range - pixel_point.y)/length_ratio));
	#endif
}

Point Coordinate_System::central_coord2pixel(Function::pPoint central_point){

	// look at the documentation to see the reasoning behind this calculation
	// TODO docu

	#ifdef SWITCHED_AXIS
		return Point((int)(range - central_point.y * length_ratio),
					 (int)(range - central_point.x * length_ratio));
	#else
		return Point(range - central_point.x * length_ratio,
					 range - central_point.y * length_ratio);
	#endif

}


void Coordinate_System::get_Coordinate_System(Mat* output){
	*output = system.clone();
}

void Coordinate_System::plot_f(vector<double>* poly, Mat* output, Scalar color){

	// pt1 and pt2 are for storing the calculated points
	Point pt1, pt2;
	double x;
	int x_pixel;

	// calculate the first x value (in central coordinates)
	// look at the for loop comment to see why it is zero
	x = (range - 0)/length_ratio;

	// first evaluate the function at x and then
	pt1 = Point( range - (Function::evaluate(poly, x)* length_ratio), 0 );

	// we go from the upper most pixel to the lowest pixel
	// essentially we draw the function in reverse
	// we go by pixel, since we can't draw finer than that anyway
	for(double y_pixel = 1; y_pixel < coordinate_length; y_pixel++){

		// calculate the x value (in central coordinates) based upon the y-pixel value
		x = (range - y_pixel)/ length_ratio;
		// evaluate y corresponding to x
		x_pixel = range - (Function::evaluate(poly, x) * length_ratio);

		// create a Point that can be used for drawing
		pt2 = Point( x_pixel, y_pixel );

		// draw a straight line between the two points
		line( *output, pt1, pt2, color, 1, 8, 0);

		// now store the pt2 as pt1, so we can draw a line from one point to the next
		// without needing to calculate double the time
		pt1 = pt2;
	}

}

void Coordinate_System::plot_f_in_range(vector<double>* poly,
										Mat* output,
										Scalar color,
										double x0, double x1){

	// for the reasoning inside this function we only need to be concerned with x
	// and f(x) and not with two polynomial functions that need the same parameter t

	// pt1 and pt2 are for storing the calculated points
	Point pt1, pt2;
	// x needs to be calculated based upon their corresponding pixel
	double x;
	int x_pixel;

	// calculate the first pixel, based upon the interval [x0, x1]
	pt1 = central_coord2pixel(Function::pPoint(x0, Function::evaluate(poly, x0)));

	// calculate the pixel representing end of the range
	int end = range - x1 * length_ratio;

	// now we go pixel by pixel to determine the corresponding x to that pixel.
	// we go by pixel, since we can't draw finer than that anyway
	// we go in the negative direction for the pixels,
	// since the smaller x-values (in central coordinates) have a higher y-value
	for(double y_pixel = pt1.y; y_pixel > end; y_pixel--){

		// since only x needs to be transformed, the conversion function is not used,
		// as it would try to calculate y as well and this would consume extra processing power
		x = (range - y_pixel)/ length_ratio;

		// first this line calculates the y-value corresponding to the x-value
		// then it calculates the pixel value corresponding to that
		x_pixel = range - (Function::evaluate(poly, x) * length_ratio);

		// create a new point (remember that the coordinate system is basically rotated
		pt2 = Point( x_pixel , y_pixel );

		// draws a straight line between pt1 and pt2
		line( *output, pt1, pt2, color, 1, 8, 0);

		// now store the pt2 as pt1, so we can draw a line from one point to the next
		// without needing to calculate double the time
		pt1 = pt2;
	}

}

void Coordinate_System::plot_vec_func(Mat* output,
									  Function::vec_func* v,
									  double start_t,
									  double end_t,
									  double granualirity,
									  Scalar color){

	// these two points are for drawing a line inside the the picture and hence will be
	// in integer pixel coordinate
	Point pt1, pt2;

	// this line first evaluates the vector function f at the point start_t and then
	// converts them into central_coordinates for storage and later use
	pt1 = central_coord2pixel(Function::evaluate_vec_func(v, start_t));

	// we go through all the following t's based on the chosen granularity
	// lets say start_t is 3 and granularity is 0.1 then the next t we will be
	// evaluating is: 3.1. The one after that is 3.2 and so on
	for(double t = start_t + granualirity; t < end_t; t += granualirity){

		// does the same as pt1
		pt2 = central_coord2pixel(Function::evaluate_vec_func(v, t));

		// draws a straight line between pt1 and pt2
		// this is the reason why the granularity is so important for a good looking graph
		// the smaller the granularity, the smoother the function will look inside
		// the coordinate system
		line( *output, pt1, pt2, color, 2, 8, 0);

		// now store the pt2 as pt1, so we can draw a line from one point to the next
		// without needing to calculate double the time
		pt1 = pt2;
	}

}

void Coordinate_System::plot_vec_func(Mat* output,
									  Function::vec_func* v,
									  Scalar color){

	plot_vec_func(output, v, v->start_t, v->end_t, 0.01, color);

}


void Coordinate_System::draw_points(Mat* output, vector<Function::pPoint>* points, Scalar color){

	// store the size outside the for loop, so the function doesn't need to be called evrytime
	// the comparison is made
	int size = points->size();

	// go through all the given points
	for(int i = 0; i < size; i++){
		// Draw a small filled circle to represent the point on the given
		// coordinate system
		circle(*output, central_coord2pixel(points->at(i)), 2, color, 2, 8, 0);
	}

}


void Coordinate_System::plot_ego_lane(Mat* output, vector<Function::vec_func>* ego_lane){

	// go through each part of the ego lane
	for(unsigned int i = 0; i < ego_lane->size(); i++){
		// and then plot them individually
		plot_vec_func(output,
					  &ego_lane->at(i),
					  ego_lane->at(i).start_t,
					  ego_lane->at(i).end_t,
					  0.01,
					  Scalar(110*i, 230, 255 - i*110));
	}
}

void Coordinate_System::plot_lane_points(Mat* output,
										 vector<vector<vector<Function::pPoint>>>* lane_vector){

	int size = lane_vector->size();
	int sub_size;

	Scalar side_color;

	for(int side = 0; side < size; side++){

		sub_size = lane_vector->at(side).size();
		side_color = Scalar(0, side*120%255, 255);

		for(int sub_line = 0; sub_line < sub_size; sub_line++){

			draw_points(output, &(lane_vector->at(side).at(sub_line)), side_color);

		}
	}
}

void Coordinate_System::plot_lane_vecs(Mat* output,
									   vector<vector<Function::vec_func>>* lane_line_vecs){

	int size = lane_line_vecs->size();
	int sub_size;

	Scalar side_color;
	Function::vec_func* v;


	for(int side = 0; side < size; side++){

		sub_size = lane_line_vecs->at(side).size();
		side_color = Scalar(0, side*120%255, 255);

		for(int sub_line = 0; sub_line < sub_size; sub_line++){

			v = &lane_line_vecs->at(side).at(sub_line);

			plot_vec_func(output,
						  v,
						  v->start_t,
						  v->end_t,
						  0.01,
						  side_color);

		}
	}
}


